/* raddic.c : kanji radical dictionary lookup via the curses screen
   manipulation library : Ben Bullock : started 22 April 1994 */

/* raddic: a curses program for finding kanji radicals */

/* Copyright (C) 1996 Ben Kasmin Bullock  */

/* raddic is free software; you can redistribute it and/or modify it under */
/* the terms of the GNU General Public License as published by the Free */
/* Software Foundation; either version 2, or (at your option) any later */
/* version. */

/* raddic is distributed in the hope that it will be useful, but WITHOUT */
/* ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or */
/* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License */
/* for more details. */

/* You should have received a copy of the GNU General Public License */
/* along with raddic; see the file COPYING.  If not, write to the Free */
/* Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA */
/*  02111-1307, USA. */



#include <curses.h>
#include <stdlib.h>
#include <stdio.h>

int movecursor();

void printyx();
void sprintyx();
void initchoicewindow();
void printrad();
void printkan();
void gettsize();

struct kanji {
char rad[2]; 
int num;     /* radical number of the kanji */
int str;  /* stroke count of the kanji */
};

struct kanji radicals[] = { /* list of all the radicals */
{"",-1,0},{"",-2,0},
{"",0 ,0 },{"",1,0},{"",2,0},{"Ц",3,0},{"",4,0},{"",5,0},{"Э",6,0},
{"",0,0},{"",7,0},{"е",8,0},{"",9,0},{"",9,0},{"Ф",9,0},{"ѹ",10,0},
{"",11,0},{"Ȭ",12,0},{"",13,0},{"",14,0},{"",15,0},{"",16,0},
{"",17,0},{"",18,0},{"",18,0},{"",19,0},{"ұ",20,0},{"Ҹ",21,0},
{"ҹ",22,0},{"",24,0},{"",25,0},{"",26,0},{"",27,0},{"",28,0},
{"",29,0},{"",162,0},{"ˮ",163,0},{"",170,0},{"",0,0},{"",30,0},
{"",31,0},{"",32,0},{"",33,0},{"",34,0},{"ͼ",36,0},{"",37,0},
{"",38,0},{"",39,0},{"",40,0},{"",41,0},{"",42,0},{"",42,0},
{"",43,0},{"",44,0},{"֥",45,0},{"",46,0},{"",47,0},{"",47,0},
{"",48,0},{"",49,0},{"̦",49,0},{"",50,0},{"",51,0},{"",52,0},
{"",53,0},{"׮",54,0},{"װ",55,0},{"׵",56,0},{"",57,0},{"",58,0},
{"",58,0},{"",58,0},{"",59,0},{"",60,0},{"˻",61,0},{"٩",64,0},
{"",85,0},{"",94,0},{"",140,0},{"",0,0},{"",61,0},{"",61,0},
{"",62,0},{"",63,0},{"",64,0},{"",65,0},{"ڽ",66,0},{"ھ",66,0},
{"ʸ",67,0},{"",68,0},{"",69,0},{"",70,0},{"",71,0},{"",71,0},
{"",72,0},{"۩",73,0},{"",74,0},{"",75,0},{"",76,0},{"",77,0},
{"",78,0},{"",79,0},{"",80,0},{"",80,0},{"",81,0},{"",82,0},
{"",83,0},{"",84,0},{"",85,0},{"",86,0},{"ۿ",86,0},{"",87,0},
{"",88,0},{"",89,0},{"",90,0},{"",91,0},{"",92,0},{"",93,0},
{"",94,0},{"",96,0},{"",113,0},{"Ϸ",125,0},{"Ͼ",130,0},{"",0,0},
{"",95,0},{"",96,0},{"",97,0},{"",98,0},{"",99,0},{"",100,0},
{"",101,0},{"",102,0},{"ɥ",103,0},{"",104,0},{"",105,0},
{"",106,0},{"",107,0},{"",108,0},{"",109,0},{"̷",110,0},
{"",111,0},{"",112,0},{"",113,0},{"",114,0},{"",115,0},
{"",116,0},{"Ω",117,0},{"",122,0},{"",145,0},{"",0,0},{"",118,0},
{"",119,0},{"",120,0},{"",121,0},{"",122,0},{"",123,0},
{"",124,0},{"Ϸ",125,0},{"",126,0},{"",127,0},{"",128,0},
{"",129,0},{"",130,0},{"",131,0},{"",132,0},{"",133,0},
{"",134,0},{"",135,0},{"",136,0},{"",137,0},{"",138,0},
{"",139,0},{"",140,0},{"",141,0},{"",142,0},{"",143,0},
{"",144,0},{"",145,0},{"",146,0},{"",0,0},{"",147,0},
{"",148,0},{"",149,0},{"ë",150,0},{"Ʀ",151,0},{"",152,0},
{"",153,0},{"",154,0},{"",155,0},{"",156,0},{"­",157,0},
{"",158,0},{"",159,0},{"",160,0},{"ä",161,0},{"͸",163,0},
{"",164,0},{"",165,0},{"Τ",166,0},{"",0,0},{"",167,0},{"Ĺ",168,0},
{"",169,0},{"",170,0},{"",171,0},{"",172,0},{"",173,0},
{"",174,0},{"",175,0},{"",210,0},{"",0,0},{"",176,0},{"",177,0},
{"",178,0},{"",179,0},{"",180,0},{"",181,0},{"",182,0},
{"",183,0},{"",184,0},{"",185,0},{"",186,0},{"10",0,0},
{"",187,0},{"",188,0},{"",189,0},{"",190,0},{"",191,0},{"",192,0},
{"",193,0},{"",194,0},{"ε",212,0},{"11",0,0},{"",195,0},{"Ļ",196,0},
{"",197,0},{"",198,0},{"",199,0},{"",200,0},{"",213,0},{"12",0,0},
{"",201,0},{"",202,0},{"",203,0},{"",204,0},
{"13",0,0},{"",205,0},{"Ť",206,0},{"",207,0},{"",208,0},
{"14",0,0},{"ɡ",209,0},{"",210,0},{"15",0,0},{"",211,0},{"16",0,0},
{"ζ",212,0},{"",213,0},{"17",0,0}, {"",214,0}};

#define NRAD 260 /* total number of radicals in the above list */

struct kanji symbols[] = { /* list of symbols */
{"",1,0}, /* ƥ */
{"",1,0}, /* ƥ */
{"",1,0}, /* ϥƥ */
{"",1,0}, /* ĥܥ, С饤 */
{"",1,0}, /* 饤 */
{"",1,0}, /* ꥫ */
{"",1,0}, /* ꥫ */
{"",1,0}, /* ꥫ */
{"",1,0}, /* ꥫ */
{"",1,0}, /* ʥ */
{"",1,0}, /* ɥ祦 */
{"",1,0}, /* ʥ, , ꥫ */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* , ʥ, ˥ */
{"",1,0}, /* ƥ, ƥƥ */
{"",1,0}, /* ƥƥ */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* å */
{"",1,0}, /* ץ饹 */
{"",1,0}, /* ޥʥ */
{"",1,0}, /* ץ饹ޥʥ */
{"",1,0}, /* , Х, ڥ */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* Υåȥ */
{"",1,0}, /* 祦ʥ */
{"",1,0}, /* ʥ */
{"",1,0}, /* , 祦ʥꥤ */
{"",1,0}, /* 祦, ʥꥤ */
{"",1,0}, /* ॲ */
{"",1,0}, /* 楨 */
{"",1,0}, /*  */
{"",1,0}, /* ᥹ */
{"",1,0}, /*  */
{"",1,0}, /* ե */
{"",1,0}, /* ӥ祦 */
{"",1,0}, /* ɥ */
{"",1,0}, /*  */
{"",1,0}, /* ɥ */
{"",1,0}, /*  */
{"",1,0}, /* ݥ */
{"",1,0}, /* ѡ */
{"",1,0}, /* ʥС */
{"",1,0}, /* ѥ */
{"",1,0}, /* ۥ */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* ۥ */
{"",1,0}, /* ۥ */
{"",1,0}, /* ޥ */
{"",1,0}, /* ޥ */
{"",1,0}, /* ˥奦ޥ, ޥ */
{"",1,0}, /* ҥ */
{"",1,0}, /* ҥ */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* 󥫥 */
{"",1,0}, /* 󥫥 */
{"",1,0}, /* 㥯󥫥 */
{"",1,0}, /* 㥯󥫥 */
{"",1,0}, /*  */
{"",1,0}, /* ᥸륷 */
{"",1,0}, /* , 楦ӥ */
{"",1,0}, /* ߥ, , 䥸륷 */
{"",1,0}, /* ҥ, 䥸륷 */
{"",1,0}, /* , 䥸륷 */
{"",1,0}, /* , 䥸륷 */
{"",1,0}, /*  */
{"",1,0}, /* ʥ */
{"",1,0}, /* ɥ */
{"",1,0}, /* ٥ */
{"",1,0}, /*  */
{"",1,0}, /* ˥ꥤ */
{"",1,0}, /* 롼 */
{"",1,0}, /* ʥʥ */
{"",1,0}, /* 󥰥ȥ */
{"",1,0}, /* ѡߥ */
{"",1,0}, /* 㡼 */
{"",1,0}, /* եå */
{"",1,0}, /*  */
{"",1,0}, /*  */
{"",1,0}, /* ֥ */
{"",1,0}, /* ѥ饰 */
{"",1,0}, /* ޥ */
{"",1,0}, /* 襳 */
{"",1,0}, /* ƥ */
{"",1,0}, /*  */
{"",1,0}, /* 襳 */
{"",1,0}  /* ƥ */
};

#define NSYMBOLS 111 /* total number of symbols in the above list */

struct kanji greek[] = {      
{"",1,0},   {"",1,0}, /* ե */     
{"",1,0},   {"",1,0}, /* ١ */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /* ǥ륿 */       
{"",1,0},   {"",1,0}, /* ץ */   
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /* å */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /* ߥ塼 */       
{"",1,0},   {"",1,0}, /* ˥塼 */       
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /* ߥ */   
{"",1,0},   {"",1,0}, /* ѥ */         
{"",1,0},   {"",1,0}, /*  */         
{"",1,0},   {"",1,0}, /*  */       
{"",1,0},   {"",1,0}, /*  */         
{"",1,0},   {"",1,0}, /* ץ */   
{"",1,0},   {"",1,0}, /* ե */       
{"",1,0},   {"",1,0}, /*  */         
{"",1,0},   {"",1,0}, /* ץ */       
{"",1,0},   {"",1,0}  /* ᥬ */       
};

#define NGREEK 48

/* char *rad_eigo[220] = { /* radical names in English *
}
char *rad_japan[220] = { /* radical names in Japanese *
} */

/* the following include statement also ends and defines
   #define KANJIS NNNN where NNNN is the number of kanjis 
   in kanjidic */

struct kanji k[] = {
#include "kanjis"

struct kanji temp[400];
struct kanji kchoice[400];
int nkchoice;

int mychar, myx, myy, oldx, oldy;

WINDOW *talk_window; /* debugging output window */

WINDOW *choicewin;
int choicex,choicey; /* x and y dimensions of the choice window */

WINDOW *titlewindow;

int radi; /* current choice */

/* the following variables are to do with the terminal size */

#define HJUMP 4 /* there must be a gap of two ascii whitespace
		   characters between the kanji for the curses display
		   to work properly at least on the Hewlett-Packard 
		   computer */

#define MINCHARS 24*80 /* the minimum number of allowable characters
			  to display on the screen */

/* width   = kanjis displayable on the screen ( = COLS/HJUMP) */

int width;

int stage;

#define RADICAL 1
#define KANJI 2

main()
{
  int i,n,t; /* counters */
  int untried;
  int now,start,end,nrad,vsize,hsize;

  /* set up the curses environment */

  initscr();
  gettsize();

  nonl();
  cbreak();
  noecho();

  initchoicewindow(radicals,NRAD);
  talk_window=newwin(2, COLS-4, LINES-3, 1);
  scrollok(talk_window, TRUE);
  titlewindow=newwin(1, 20, 0, COLS/2-5);
  scrollok(titlewindow, TRUE);
  wprintw(titlewindow,"\nRADICALS");
  wrefresh(titlewindow);

  /* put cursor on a correct character to start off with */

  /* mychar = KEY_RIGHT; 
     radi = movecursor(NRAD); */

  printrad(radi,1);

  /* the following variable is needed in the movecursor routine : it
     tells which of kanji or radical stages we are in */

  stage = RADICAL;

  while((mychar=wgetch(choicewin))!='q' || radicals[radi].num == 0) {

    printrad(radi,0);
    radi = movecursor(NRAD);
    printrad(radi,1);
  }

  delwin(choicewin);
  clear();
  refresh();

  if ( radicals[radi].num == -1 ) {
    wprintw(titlewindow,"\nSYMBOLS");
    wrefresh(titlewindow);
    initchoicewindow(symbols,NSYMBOLS);
    endwin();
    exit(0);
  }
    
  if ( radicals[radi].num == -2 ) {
    wprintw(titlewindow,"\nGREEK");
    wrefresh(titlewindow);
    initchoicewindow(greek,NGREEK);
    endwin();
    exit(0);
  }
    
  stage = KANJI;

  t=0;
  for (i=0;i<KANJIS;i++) {
    if (k[i].num==radicals[radi].num) {
      temp[t]=k[i];
      t++;
    }
  }
  
  wprintw(titlewindow,"\nRADICAL %c%c",
	  radicals[radi].rad[0],radicals[radi].rad[1]);
  wrefresh(titlewindow);

  if (t==0) { /* this is testing for a radical with no kanjis associated with
		 it ? */
    endwin();
    fprintf(stderr," No kanji at all!\n");
    exit(0);
  }

  /* now sort all the kanji */

  nkchoice=0;

  for (i=1;i<31;i++) { /* counting number of strokes of kanji */
    untried=0;
    for (n=0;n<t;n++) {
      if (temp[n].str==i) {

	if (untried==0) {

	  untried = 1;

	  /* put a number into this slot */

	  if (i<10) {
	    kchoice[nkchoice].rad[0]=' ';
	    kchoice[nkchoice].rad[1]='0'+i;
	  } else {
	    kchoice[nkchoice].rad[1]='0'+i%10;
	    kchoice[nkchoice].rad[0]='0'+i/10;
	  }
	  kchoice[nkchoice].num=0;
	  kchoice[nkchoice].str=0;

	  nkchoice++;
	}

	kchoice[nkchoice]=temp[n];

	nkchoice++;
      }
    }
  }

  initchoicewindow(kchoice,nkchoice);

/* the following piece of code puts the cursor on a "character" to
   start with: at least for the kanji part I found that this is not
   useful since there are too many number boxes that it can be
   difficult to get onto the right kanji without contortions.
  
  mychar = KEY_RIGHT;
  radi = movecursor(nkchoice);
  printkan(kchoice,radi,1); */
  
  while((mychar=wgetch(choicewin))!='q') {
    printkan(kchoice,radi,0);
    radi = movecursor(nkchoice);
    printkan(kchoice,radi,1);
    if (kchoice[0].str == 0) {
      wprintw(talk_window,"\n%c%c strokes: %d",
	      kchoice[radi].rad[0],kchoice[radi].rad[1],kchoice[radi].str);
      wrefresh(talk_window);
    }
  }
  endwin();
  exit(0);
}

int movecursor(maxnow) 

/* move the cursor around the screen subject to it still being on a
   character */

int maxnow;
{
  int maxy,maxx;
  int now;

  maxx=choicex*HJUMP;
  maxy=choicey;

 again:

  oldx=myx;
  oldy=myy;
  
  if (mychar == KEY_RIGHT) {
    myx += HJUMP;
    if (myx > maxx-HJUMP) myx = 0;
    now = myx/HJUMP+myy*maxx/HJUMP;
    if (now >= maxnow) {
      myx = 0;
    }
  }
  if (mychar == KEY_LEFT) {
    myx -= HJUMP;
    if (myx < 0) myx = maxx-HJUMP;
  }
  if (mychar == KEY_DOWN) {
    myy++;
    now = myx/HJUMP+myy*maxx/HJUMP;
    if (myy > maxy - 1 || (now >= maxnow || now < 0) ) {
      myy = 0;
    }
  }
  if (mychar == KEY_UP) {
    myy--;
    if (myy < 0 ) {
      myy = maxy-1;
      now = myx/HJUMP+myy*maxx/HJUMP;
      if (myy > maxy - 1 || (now >= maxnow || now < 0) ) {
	myy--;
      }
    }
  }

  now = myx/HJUMP+myy*maxx/HJUMP;

  if (now >= maxnow || now < 0) { /* the cursor is no longer on a character */
    myx=oldx;
    myy=oldy;
    now=myx/HJUMP+myy*maxx/HJUMP;
  }
  wprintw(talk_window,"%d %d\n", myx, myy);
  wrefresh(talk_window);

  /* Now test that we are on a genuine radical - if not then just go
     back to the beginning of this routine and move the cursor again.

     You are advised not to use a "goto" statement in your program,
     but I don't know why.  It's the easiest way to do things like the
     following. */

  if (stage == RADICAL && radicals[now].num == 0) goto again;

/* add the following to the if to make it jump over numbers in the kanji part ||
      (stage == KANJI && kchoice[now].str == 0)) */

  return(now);
}

void printyx(kan)

/* print a kanji at the specified location */

struct kanji kan;

{
  wmove(choicewin,myy,myx);
  wprintw(choicewin," %c%c ",kan.rad[0],kan.rad[1]);
  wrefresh(choicewin);
}

void sprintyx(kan)

/* print a standing out kanji at the specified location */

struct kanji kan;

{
  wattrset(choicewin,A_STANDOUT);
  wmove(choicewin,myy,myx);
  wprintw(choicewin," %c%c ",kan.rad[0],kan.rad[1]);
  wrefresh(choicewin);
  wattroff(choicewin,A_STANDOUT);
}

void printrad(radnum,onoff)

/* print all radicals out with the same number as the current one */

int radnum,onoff;

{
  int i;
  
  if (radicals[radnum].num==0) {
    if (onoff) {
      printyx(radicals[radnum]);
    } else {
      sprintyx(radicals[radnum]);
    } 
  } else {

    if (onoff) wattrset(choicewin,A_STANDOUT);
    
    for (i=0;i<NRAD;i++) {
      if (radicals[i].num==radicals[radnum].num) {
	wmove(choicewin,i/choicex,HJUMP*(i%choicex));
	wprintw(choicewin," %c%c ",radicals[i].rad[0],radicals[i].rad[1]);
      }
    } /* end of loop over all radicals */
    
    if (onoff) wattroff(choicewin,A_STANDOUT);
    wrefresh(choicewin);
  }
}

void printkan(kchoice,kannum,onoff)

struct kanji kchoice[];
int kannum,onoff;

{
  if (kchoice[kannum].num==0) {
    if (onoff) {
      printyx(kchoice[kannum]);
    } else {
      sprintyx(kchoice[kannum]);
    } 
  } else {
    if (onoff) {
      sprintyx(kchoice[kannum]);
    } else {
      printyx(kchoice[kannum]);
    } 
  }
  wrefresh(choicewin);
}

void initchoicewindow(choice,nchoice)

     /* make the window for choosing a particular character, and move the
	cursor around inside the window until we find the one that we want */

     struct kanji choice[]; /* this contains the list of elements to 
			       choose from */
     int nchoice; /* number of elements in choice */

{

  int i; /* general purpose counter */
  
  /* first set global variables to zero */

  myy=0;
  myx=0;
  radi=0;

  if (nchoice < width) {
    choicex=nchoice;
    choicey=1;
  } else {
    choicex=width;
    choicey=nchoice/width+1;
  }

  if ((choicewin=newwin(choicey,choicex*HJUMP, 2, 1)) == NULL) {
    fprintf(stderr,"Raddic could not open a window.\n");
    exit(1);
  }

  keypad(choicewin,1); /* so that the cursor keys can be used */

  wmove(choicewin,0,0);

  /* print out the initial screen */

  for (i=0;i<nchoice;i++) {
    wmove(choicewin,i/choicex,(i%choicex)*HJUMP);
    if (choice[i].num==0) {
      wattrset(choicewin,A_STANDOUT);
      wprintw(choicewin," %c%c ",choice[i].rad[0],choice[i].rad[1]);
      wattroff(choicewin,A_STANDOUT);
    } else {
      wprintw(choicewin," %c%c ",choice[i].rad[0],choice[i].rad[1]);
    }
  }

  wrefresh(choicewin);
}

void gettsize()  /* get the size of the terminal window */

{
  width = (COLS-4)/HJUMP;

  if ( LINES*COLS < MINCHARS ) {
    endwin();
    printf("Your terminal size of %d X %d characters ",LINES,COLS);  
    printf("is too small.\nPlease increase your terminal ");
    printf("size to allow %d characters",MINCHARS);
    printf(" characters.\n");
    exit(1);
  }
}

/* planned additions to this program:

(1) toggle English language / Japanese language instructions
(2) link into xjdic program so that can use the kanjidic indices
(3) make blow-ups of kanji using the kbanner program
(4)

*/

